     
     ; some device types...
    .equiv		DEVICE_IS_24F,				0xA1
    .equiv		DEVICE_IS_24FJ,				0xA2
    .equiv		DEVICE_IS_24HJ,				0xA3
    .equiv		DEVICE_IS_24E,				0xA4
    .equiv		DEVICE_IS_30F,				0xA5
    .equiv		DEVICE_IS_33FJ,				0xAA
    .equiv		DEVICE_IS_33E,				0xAC
 
    ; ==============================================================
    ; CHANGE
    ; ==============================================================
    ; select device...
	.include	"p24HJ128GP502.inc"
	config	__FWDT,		FWDTEN_OFF & WINDIS_OFF		
    config __FOSCSEL, FNOSC_FRCPLL & IESO_OFF 
    config __FOSC, FCKSM_CSECMD & OSCIOFNC_ON & POSCMD_NONE & IOL1WAY_OFF

    ; clock and communication baudrates...
    .equiv 		DEVICE_CLOCK,               32256000			; clock (32.256)
    .equiv		BAUDRATE,                 	38400				; baudrate

    ; DATASHEET NOTES
    ; With RTSP, the user application can write program memory data either in
    ; block or rows of 64 instructions (192 bytes) at a time or a single program 
    ; memory word, and erase program memory in blocks or pages of 512 
    ; instructions (1536 bytes) at a time.
    ; *** 24FJ128GA010
    ;.equiv      DEVICE_ID,                	0x0216				; used by PC application only (optional) 
    ;.equiv      DEVICE_SERIES,            	DEVICE_IS_24FJ		; used by PC and this application (mandatory)
    ;.equiv      DEVICE_ROM,               	0x15800				; in words!!
    ;.equiv      DEVICE_BLOCK_ERASE,       	1536				; bytes, 8 pages * 192
    ;.equiv      DEVICE_BLOCK_WRITE,       	192					; bytes
    ;.equiv      DEVICE_EEPROM,            	0					; no onboard EEPROM

    ; *** 24HJ128GP502
    .equiv      DEVICE_ID,                	0x67D0				; used by PC application only (optional) 
    .equiv      DEVICE_SERIES,            	DEVICE_IS_24HJ		; used by PC and this application (mandatory)
    .equiv      DEVICE_ROM,               	0x15800				; in words!!
    .equiv      DEVICE_BLOCK_ERASE,       	1536				; bytes, 8 pages * 192
    .equiv      DEVICE_BLOCK_WRITE,       	192					; bytes
    .equiv      DEVICE_EEPROM,            	0					; no onboard EEPROM

    ; FLASH and EEPROM erase and write - may change depending
    ; on device series...
	.equiv 		ROM_ERASE,             		0x4042				; FJ, HJ series	- tested djb 
	.equiv 		ROM_WRITE,             		0x4001				; FJ, HJ series - tested djb

	;.equiv 	ROM_ERASE,             		0x4058				; F series 
	;.equiv 	ROM_WRITE,             		0x4004				; F series
	;.equiv 	EEPROM_ERASE,          		0x4058				; F series
	;.equiv 	EEPROM_WRITE,          		0x4004 				; F series

	;.equiv 	ROM_ERASE,					0x4041				; 30F series
	;.equiv 	ROM_WRITE,					0x4001				; 30F series
	;.equiv 	EEPROM_ERASE,				0x4044				; 30F series
	;.equiv 	EEPROM_WRITE,				0x4004				; 30F series
    ; ==============================================================
    ; END CHANGE
    ; ==============================================================

	; formula is (OSC / BAUDRATE / MULT) - 1, we need to round the result
	; so that the SPBRG value is correct
	.equiv      DEVICE_CLOCK_FCY,			 DEVICE_CLOCK / 2

    ; BRGH = 0
	;.equiv		SPBRG_VALUE,                 (DEVICE_CLOCK_FCY / BAUDRATE + 8) / 16 - 1
    
    ; BRGH = 1
	.equiv		SPBRG_VALUE,                 (DEVICE_CLOCK_FCY / BAUDRATE + 8) / 4 - 1

    ; alias UART registers...
   	.equ    	_umode, 					 U1MODE	
	.equ        _usta,  	 				 U1STA			
	.equ        _ubrg,						 U1BRG				
    .equ		_rcReg, 					 U1RXREG
    .equ		_txReg,						 U1TXREG

     ; timing
	.equiv		TIMEOUT_MS,					 2000											   ; timeout in ms
	.equiv		TIMEOUT,					 TIMEOUT_MS * (DEVICE_CLOCK / 1000) / (65536 * 14) ; number of times to perform inner loops

	; set the loader size and user loader origin...
	.equiv 		LOADER_SIZE,           		DEVICE_BLOCK_ERASE / 3 * 2 
	.equiv 		ORG_LOADER,		            DEVICE_ROM - LOADER_SIZE
    .equiv 		ORG_LOADER_USER,           	ORG_LOADER - 4 		; words

	; loader commands - these are bit field identifiers
	; data input values are shown in comments
	.equiv 		CMD_EEPROM_READ,           	0x00              	; 0x01
	.equiv 		CMD_EEPROM_WRITE,          	0x01              	; 0x02
	.equiv 		CMD_ROM_ERASE,             	0x02              	; 0x04
	.equiv 		CMD_ROM_READ,              	0x03              	; 0x08
	.equiv 		CMD_ROM_WRITE,             	0x04              	; 0x10
	.equiv 		CMD_BAUDRATE,              	0x05              	; 0x20

	; addition information and commands...
	.equiv 		FIRMWARE_VERSION,          	0x10
	.equiv 		LOADER_CONNECT,            	0x80
	.equiv 		LOADER_ACK,                	LOADER_CONNECT
	.equiv 		LOADER_GO,                 	0x81
	.equiv 		LOADER_ERROR_CRC,          	0x8A
	.equiv 		LOADER_ERROR_VERIFY,       	0x8B

	 ; program variables
	.global 	Buffer
	.equiv		BUFFER_SIZE,				DEVICE_BLOCK_WRITE
	.bss		Buffer, BUFFER_SIZE, 8	
    .equiv		Addr,						W2
    .equiv		AddrVerify,					W3
    .equiv  	Pointer,					W4
    .equiv		CRC,						W5
	.equiv		Command,					W6
	.equiv		Ack,						W7
	.equiv		Index,						W8
	.equiv		Timeout_01,					W9
    .equiv		Timeout_02,					W10
	.equiv		StackPtr,					W15
			
    ; relocated reset vector - just below loader origin...
    .section *, code, address(ORG_LOADER_USER)
LoaderStartUser:	
    nopr
	nopr
		
    ; start of loader
    .global __reset
	.section *, code, address(ORG_LOADER)
__reset:
   	mov			#__SP_init, StackPtr

    ; ==============================================================
    ; CHANGE
    ; ==============================================================
    ; configure the PLL - default values are
    ; PLLPRE = 2
    ; PLLDIV = 50
    ; PLLPOST = 4
    ; for example, assuming OSC = 7.37 (nom) then
    ; 7.37 / 2 * 50 = 184.25MHz / 4 = 46.0625MHz
    ; we don't want to use OSCTUN, as from the datasheet:
    ;    "OSCTUN functionality has been provided to help customers compensate 
    ;     for temperature effects on the FRC frequency over a wide range of 
    ;     temperatures. The tuning step size is an approximation and is neither
    ;     characterized nor tested."
    ; if we set PLLDIV = 35, then OSC = 32.24375
    ; IMPORTANT - values are
    ; 0000 = 2
    ; 0001 = 3 and so on - remember to subract 2 from calculated value!!
    mov.w		#33, W0
    mov.w       W0, PLLFBD									; PLLDIV

	; receive, map pin (RP4 = 00100) to uart
	bclr		RPINR18, #U1RXR0		
	bclr		RPINR18, #U1RXR1		
	bset		RPINR18, #U1RXR2		
	bclr		RPINR18, #U1RXR3		
	bclr		RPINR18, #U1RXR4		
						
    ; U1TX 00011 RPn tied to UART1 Transmit
	; transmit, map pin (RP5 = 00011) to uart 
	bset		RPOR2, #RP5R0			
	bset		RPOR2, #RP5R1			
	bclr		RPOR2, #RP5R2			
	bclr		RPOR2, #RP5R3			
	bclr		RPOR2, #RP5R4			
    ; ==============================================================
    ; END CHANGE
    ; ==============================================================

    ; configure uart
    mov.w 		#SPBRG_VALUE, W0				        	; move default baudrate
    mov.w 		W0, _ubrg									; into brg
    ;mov.w 		#0x8800, W0									; set the uart
    mov.w 		#0x8808, W0									; set the uart (BRGH = 1)
    mov.w 		W0, _umode									; mode
    bset  		_usta, #10                  				; transmit enable bit

    ; can we connect to remote PC...        
    rcall  		ReadByte                					; read a byte
    sub  		#LOADER_CONNECT, W0         				; subtract connection ID                    
    bra         NZ, LoaderStartUser         				; if not zero, run user code
   
    ; acknowledge...
    mov.w  		#LOADER_CONNECT, W0							; set acknowledge
    rcall  		WriteByte									; and send...

   ; wait for go before sending device information...
WaitForGo:
    rcall 		ReadByte									; get value
    sub 		#LOADER_GO, W0								; and check fo GO
    bra   		NZ, WaitForGo                               ; loop back
		     
    ; firmware version...
    mov.w  		#FIRMWARE_VERSION, W0                       ; send
    rcall  		WriteByte                                   ; firmware version...

    ; baudrate...
    mov.w  		#((BAUDRATE >> 16) & 0xFF), W0              ; send
    rcall  		WriteByte                                   ; default baudrate UB
    mov.w  		#((BAUDRATE >> 8) & 0xFF), W0               ; send
    rcall  		WriteByte                                   ; default baudrate HB
    mov.w       #(BAUDRATE & 0xFF), W0                      ; send
    rcall  		WriteByte                                   ; default baudrate LB

    ; device series...
    mov.w       #(DEVICE_SERIES & 0xFF), W0                 ; send
    rcall  		WriteByte                                   ; device type (24F, 24FJ etc)
   
    ; device ID...
    mov.w  		#((DEVICE_ID >> 8) & 0xFF), W0              ; send
    rcall  		WriteByte                                   ; device ID HB
    mov.w  		#(DEVICE_ID & 0xFF), W0						; send
    rcall  		WriteByte                                   ; device ID LB

    ; the loader origin - this defines the loader protection point - note
    ; that the PC application will offset this value by 8 bytes to take 
    ; into account the relocatable user code address
    mov.w  		#((ORG_LOADER >> 16) & 0xFF), W0            ; send
    rcall  		WriteByte                                   ; origin UB
    mov.w  		#((ORG_LOADER >> 8) & 0xFF), W0             ; send
    rcall  		WriteByte                                   ; origin HB
    mov.w  		#(ORG_LOADER & 0xFF), W0                    ; send 
    rcall  		WriteByte                                   ; origin LB

    ; erase block size...
    mov.w  		#((DEVICE_BLOCK_ERASE >> 8) & 0xFF), W0     ; send
    rcall  		WriteByte                                   ; erase block HB
    mov.w  		#(DEVICE_BLOCK_ERASE & 0xFF), W0            ; send
    rcall  		WriteByte                                   ; erase block LB

    ; write block size...
    mov.w  		#(DEVICE_BLOCK_WRITE & 0xFF), W0			; send
    rcall  		WriteByte                                   ; write block
     
    ; init ack
    mov.w  		#LOADER_ACK, Ack                            ; initalise ack...

; main program loop
LoaderMain:
   
    ; get command...
    rcall  		ReadByte                					; read a byte
    mov.w  		W0, Command                					; set command
    clr   		CRC                     					; clear CRC
   
    ; read upper byte of address...
    rcall  		ReadByte           							; read a byte
    mov.w		W0, TBLPAG									; into table page
  
    ; read high byte of address...
    rcall  		ReadByte                					; read a byte
    mov.b		WREG, Buffer + 1                            ; move into table pointer high
   
    ; read low byte of address...
    rcall  		ReadByte           							; read a byte
    mov.b       WREG, Buffer                                ; move into table pointer low
    
    ; save copy of address for verfication
    mov.w       Buffer, Addr								; move buffer.word to address register
    mov.w       Addr, AddrVerify                            ; and address verify
       
    ; EEPROM write...
    .if DEVICE_EEPROM > 0
    btsc   		Command, #CMD_EEPROM_WRITE					; check command
    rcall  		EEPROMWrite             					; execute EEPROM write
    .endif    

    ; code erase...
    btsc  		Command, #CMD_ROM_ERASE  					; check command
    rcall  		ROMErase                					; execute program code erase 

    ; code read...
    btsc  		Command, #CMD_ROM_READ   					; check command
    rcall  		ROMRead                 					; execute program code read
   
    ; code write...
    btsc  		Command, #CMD_ROM_WRITE  					; check command
    rcall  		ROMWrite                					; execute program code write

    ; write ack...
    mov.w   	Ack, W0                  					; ack byte to w reg
    rcall  		WriteByte               					; send
    bra    		LoaderMain              					; then loop back

; verify error...
VerifyError:
   mov.w  		#LOADER_ERROR_VERIFY, W0					; set flag
   rcall  		WriteByte               					; send ack
   bra    		LoaderMain              					; loop back to main

; CRC error...
CRCError:
   mov.w  		#LOADER_ERROR_CRC, W0        				; set flag
   rcall 		WriteByte               					; send ack
   bra    		LoaderMain              					; loop back to main   

; execute user code...
RunUserCode:
   bra    LoaderStartUser         							; execute user code

;*****************************************************************************
; Name    : ReadByte                                                         *
; Purpose : Read a byte from the hardware USART. Return in WREG              *
;         : If a timeout occurs, the routine will branch to 'RunUserCode'    *
;*****************************************************************************
ReadByte:
	mov 		#TIMEOUT, Timeout_01						; set delay
TimeoutLoop_01:
    clr			Timeout_02									; clear timeout register
TimeoutLoop_02:		
	clrwdt													; clear WDT
	btss		_usta, #0									; URXDA (1 = have data)
	bra 		NotReceived									; no data, loop back
	mov 		_rcReg, W0									; have data, move to WREG
	add 		CRC, W0, CRC								; and add to CRC	
	return													; data received, return

NotReceived:	
    dec			Timeout_02, Timeout_02						; dec timeout register
    bra			nz, TimeoutLoop_02							; loop back if not zero
    dec			Timeout_01, Timeout_01						; dec timeout register	
    bra			nz, TimeoutLoop_01							; loop back if not zero
    bra			RunUserCode									; timeout - abort loader

;*****************************************************************************
; Name    : WriteByte                                                        *
; Purpose : Write a byte to the hardware USART                               *
;*****************************************************************************
WriteByte:
    clrwdt                         							; clear WDT 
	btss		_usta, #8									; TRMT (0 = busy)
	bra			WriteByte									; loop back
	mov 		W0, _txReg									; send data
	return

;*****************************************************************************
; Name    : EEPROMWrite                                                      *
; Purpose : Write a single byte of data to onboard EEPROM                    *
;*****************************************************************************
.if DEVICE_EEPROM > 0
EEPROMWrite:

    ; read data
    mov.w  		#Buffer, Pointer    						; load buffer address  
    rcall  		ReadByte           							; read a byte
    mov.b		W0, [Pointer++]								; place in buffer
    rcall  		ReadByte           							; read a byte
    mov.b		W0, [Pointer++]								; place in buffer

    ; CRC check...
    rcall  		ReadByte                					; read a byte
	cp0.b 		CRC											; test for zero
	bra 		NZ, CRCError								; not zero, so error
   
    ; Set up NVMCON for erase operation
	tblwtl 		Addr, [Addr] 								; set base address of erase block
    mov.w		#EEPROM_ERASE, W0							; move erase code			
	mov.w		W0, NVMCON 									; to flash memory control register
    rcall       WriteSequence

	; write data word...
    mov.w  		#Buffer, Pointer    						; load buffer address  
	tblwtl 		[Pointer], [Addr]							; write word
	mov.w 		#EEPROM_WRITE, W0							; move write to eeprom
	mov.w       W0, NVMCON									; to eeprom memory control register
	rcall 		WriteSequence                               ; perform required write sequence
	
    ; verify
	tblrdl		[AddrVerify], W0
	cp.b		W0, [Pointer]
	bra			NZ, VerifyError
    return
.endif ; DEVICE_EEPROM
     
;*****************************************************************************
; Name    : ROMErase                                                         *
; Purpose : Erase a block of device ROM                                      *
;         : (a) Set the NVMOP bits (NVMCOM<3:0>) to 0010 to configure for  *
;         : block erase. Set the ERASE (NVMCOM<6>) and WREN (NVMCOM<14>)     *
;         : bits.                                                            *
;         : (b) Write the starting address of the block to be erased into    *
;         : the TBLPAG and W registers.										 *
;         : (c) Write 55h to NVMKEY.										 *
;         : (d) Write AAh to NVMKEY.										 *
;         : (e) Set the WR bit (NVMCOM<15>). The erase cycle begins and the  *
;         : CPU stalls for the duration of the erase cycle. When the erase   *
;         : is done, the WR bit is cleared automatically					 *
;*****************************************************************************
ROMErase:
    ; Set up NVMCON for block erase operation
	tblwtl 		Addr, [Addr] 								; set base address of erase block
    mov.w		#ROM_ERASE, W0 								; move erase code			
	mov.w		W0, NVMCON 									; to flash memory control register

    ; init pointer to row to be ERASED - TBLPAG already initialised
    ; in main program loop
    rcall       WriteSequence								; perform required write sequence
    return

;*****************************************************************************
; Name    : ROMRead                                                          *
; Purpose : Read a single byte from ROM                                      *
;         : This is typically used to get the reset vector                   *
;*****************************************************************************
ROMRead:
    tblrdh.b	[Addr], W0
    rcall       WriteByte
	tblrdl.b 	[Addr++], W0
	rcall       WriteByte
	tblrdl.b 	[Addr++], W0
    rcall  		WriteByte 								
    return

;*****************************************************************************
; Name    : ROMWrite                                                         *
; Purpose : Write data held in data buffer to ROM                            *
;*****************************************************************************
ROMWrite:
    mov.w  		#Buffer, Pointer    						; load buffer address  
    mov.w  		#BUFFER_SIZE, Index				            ; load buffer size

    ; read all data into buffer...
ReadData:
    rcall  		ReadByte           							; read a byte
    mov.b		W0, [Pointer++]								; place in buffer
    dec			Index, Index								; dec index
    bra 		nz, ReadData								; loop back if not zero

    ; CRC check...
    rcall  		ReadByte                					; read a byte
	cp0.b 		CRC											; test for zero
	bra 		NZ, CRCError								; not zero, so error
   
    ; write the data into holding registers...
    mov.w  		#Buffer, Pointer    						; load buffer address
	mov.w 		#(BUFFER_SIZE / 3), Index					; load buffer size (in program words)

LoadData:
	tblwth.b 	[Pointer++], [Addr] 						
	tblwtl.b	[Pointer++], [Addr++]
	tblwtl.b	[Pointer++], [Addr++] 
	dec 		Index, Index
	bra 		NZ, LoadData
	
    ; write a single row
	mov.w 		#ROM_WRITE, W0								; move write to rom code
	mov.w       W0, NVMCON									; to flash memory control register
	rcall 		WriteSequence                               ; perform required write sequence

    ; now verify the data...
    mov.w  		#Buffer, Pointer    						; load buffer address
	mov.w 		#(BUFFER_SIZE / 3), Index					; load buffer size in program words

VerifyData:
    tblrdh.b 	[AddrVerify], W0							
    cp.b		W0, [Pointer++]
    bra			NZ, VerifyError

    ; Verify low byte
	tblrdl.b 	[AddrVerify++], W0
	cp.b		W0, [Pointer++]
	bra			NZ, VerifyError

	; Verify high byte
	tblrdl.b 	[AddrVerify++], W0
	cp.b		W0, [Pointer++]
	bra			NZ, VerifyError

	dec			Index, Index								; dec index
	bra 		NZ, VerifyData								; and loop back if not zero
	return
     
;*****************************************************************************
; Name    : WriteSequence                                                    *
; Purpose : A required write sequence for writes                             *
;*****************************************************************************
WriteSequence:
	mov 			#0x55, W0
	mov 			W0, NVMKEY
	mov 			#0xAA, W0
	mov 			W0, NVMKEY
	bset 			NVMCON, #WR
	nop
	nop	
    
    ;wait...
WaitForWrite:
	btsc			NVMCON, #WR		
	bra 			WaitForWrite				
	return   
 

      

